#include <iostream>
#include <fstream>
#include <sstream>
#include <igstkAuroraTracker.h>
#include <igstkPolarisTracker.h>
#include "DataAcquisitionSettingsFileReader.h"

#ifdef WIN32 //running on a windows system
  #include <igstkSerialCommunicationForWindows.h>
#else //running on a unix system
  #include <igstkSerialCommunicationForPosix.h>
#endif

#include <igstkTransform.h>

#include "DataAcquisition.h"

const igstk::Tracker::TimePeriodType DataAcquisition::DEFAULT_VALIDITY_TIME = 400;

DataAcquisition::DataAcquisition(DataAcquisitionSettings &settings) : stopAcquisition(false),                                     
                                                                      tracking(false),
                                                                      outputFiles(NULL), 
                                                                      numAcquiredSamples(NULL),
                                                                      validityTime(DataAcquisition::DEFAULT_VALIDITY_TIME)
{
  this->errorObserver = IGSTKErrorObserver::New();

#ifdef WIN32 //running on a windows system
  this->serialCommunication = igstk::SerialCommunicationForWindows::New();
#else //running on a unix system
  this->serialCommunication = igstk::SerialCommunicationForPosix::New();
#endif

  SerialCommunicationSettings scSettings = settings.GetSerialCommunicationSettings();

                       //setup the serial communication
  this->serialCommunication->SetPortNumber(scSettings.GetCOMPort());
  this->serialCommunication->SetParity(scSettings.GetParity());
  this->serialCommunication->SetBaudRate(scSettings.GetBaudRate());
  this->serialCommunication->SetDataBits(scSettings.GetDataBits());
  this->serialCommunication->SetStopBits(scSettings.GetStopBits());
  this->serialCommunication->SetHardwareHandshake(scSettings.GetHandshake());

                             //get tool info  
  settings.GetToolSettings(this->trackerTools);

                            //get acquisition settings
  this->sampleNumber = settings.GetSampleNumber();
  this->currentFileNumber = settings.GetInitialFileNumber();

                       //create auxilliary information used during acquisition        
  unsigned int numberOfTools = this->trackerTools.size();
  this->outputFiles = new std::pair<std::string, std::ofstream>[numberOfTools];
  this->numAcquiredSamples = new int[numberOfTools];
  this->sampleSize = numberOfTools*this->sampleNumber;

                       //create and setup tracker
  igstk::AuroraTracker::Pointer auroraTracker;
  igstk::PolarisTracker::Pointer polarisTracker;

  std::vector<NDIToolSettings *>::iterator it = this->trackerTools.begin(); 

  if(settings.GetTrackingSystemType() == TrackingSystemsEnum::POLARIS_VICRA_VALUE) {
    this->tracker = polarisTracker = igstk::PolarisTracker::New();
    for(; it!=this->trackerTools.end(); it++) {
        if(!(*it)->GetSROMFileName().empty())  //polaris only has ports and no channels (fixed channel=0)      
          polarisTracker->AttachSROMFileNameToPort((*it)->GetPort(), (*it)->GetSROMFileName());
    }
    polarisTracker->SetCommunication(this->serialCommunication);
  }
  else if(settings.GetTrackingSystemType() == TrackingSystemsEnum::AURORA_VALUE) {
    this->tracker = auroraTracker = igstk::AuroraTracker::New();
    for(; it!=this->trackerTools.end(); it++) {
        if(!(*it)->GetSROMFileName().empty())        
          auroraTracker->AttachSROMFileNameToPort((*it)->GetPort(), (*it)->GetSROMFileName());
    }
    auroraTracker->SetCommunication(this->serialCommunication);
  }
           //how long is a transformation valid (in milliseconds)
  this->tracker->SetValidityTime(this->validityTime);

           //observe the serial communication and tracker for errors
  this->serialCommunication->AddObserver(igstk::OpenPortErrorEvent(), this->errorObserver);
  this->serialCommunication->AddObserver(igstk::ClosePortErrorEvent(), this->errorObserver);
  
  this->tracker->AddObserver(igstk::TrackerOpenErrorEvent(), this->errorObserver);
  this->tracker->AddObserver(igstk::TrackerInitializeErrorEvent(), this->errorObserver);
  this->tracker->AddObserver(igstk::TrackerStartTrackingErrorEvent(), this->errorObserver);
  this->tracker->AddObserver(igstk::TrackerStopTrackingErrorEvent(), this->errorObserver);
  this->tracker->AddObserver(igstk::TrackerUpdateStatusErrorEvent(), this->errorObserver);
  this->tracker->AddObserver(igstk::TrackerCloseErrorEvent(), this->errorObserver);
  this->tracker->AddObserver(igstk::TrackerUpdateStatusErrorEvent(), this->errorObserver);
}
/******************************************************************************/
void DataAcquisition::Delete()
{
           //stop tracking, close tracker communication, and close com port
           //I don't check if an IGSTK error occured as I can't do a thing about it
  if(this->tracking) {          
    this->tracker->RequestStopTracking();            
    this->tracker->RequestClose();  
    this->serialCommunication->CloseCommunication();
  }  
         //close all output streams
  if(this->outputFiles!=NULL) {
    unsigned int numberOfTools = this->trackerTools.size();
    for(int i=0; i<numberOfTools; i++)
      this->outputFiles[i].second.close();
  }
  delete [] outputFiles;
  delete [] numAcquiredSamples;
  itk::Object::Delete();
}
/******************************************************************************/
void DataAcquisition::StartTracking() throw (TrackerCommunicationException)
{
  std::string errorMessage;          
            //open serial communication
  this->serialCommunication->OpenCommunication();  
  if(this->errorObserver->Error()) {
    this->errorObserver->GetErrorMessage(errorMessage);
    this->errorObserver->ClearError();
    throw TrackerCommunicationException(errorMessage);
  }

                 //open tracker communication
  this->tracker->RequestOpen();
  if(this->errorObserver->Error()) {
    this->errorObserver->GetErrorMessage(errorMessage);
    this->errorObserver->ClearError();
    this->serialCommunication->CloseCommunication();
    throw TrackerCommunicationException(errorMessage);
  }

          //initialize tracker
  this->tracker->RequestInitialize();
  if(this->errorObserver->Error()) {
    this->errorObserver->GetErrorMessage(errorMessage);
    this->errorObserver->ClearError();
    this->tracker->RequestClose();
    this->serialCommunication->CloseCommunication();
    throw TrackerCommunicationException(errorMessage);
  }

       //start tracking
  this->tracker->RequestStartTracking();
  if(this->errorObserver->Error()) {
    this->errorObserver->GetErrorMessage(errorMessage);
    this->errorObserver->ClearError();
    this->tracker->RequestClose();
    this->serialCommunication->CloseCommunication();
    throw TrackerCommunicationException(errorMessage);
  }
  this->tracking = true;
}
/******************************************************************************/
void DataAcquisition::AcquireData() throw (DataAcquisitionException)
{
  igstk::Transform currentTransform;               
  if(!this->tracking)
    throw(DataAcquisitionException("Cannot acquire data before tracking has started."));

  unsigned int i, j;
  int currentSampleSize, currentSampleIncrement;
  std::ostringstream fileNameStream, displayStream;
  std::string errorMessage;

  this->stopAcquisition = false;
  this->InvokeEvent(DataAcquisitionStartEvent()); //nofity all observers that acquisition is starting

  unsigned int numberOfTools = this->trackerTools.size();  

             //the following lines are part of the hack that allows us to check if a transformation is valid 
             //based on its time stamps
  igstk::TimeStamp::TimePeriodType currentTime; 
  igstk::TimeStamp::TimePeriodType *prevTimeStamps = new igstk::TimeStamp::TimePeriodType[numberOfTools];
  memset(prevTimeStamps,'0',numberOfTools*sizeof(igstk::TimeStamp::TimePeriodType));

             //create all the output streams, file name format is "port_a_[0,1].b.txt" where "a" is
             //the port number (starting at 1), then the channel number zero or one, and "b" is the currentFileNumber
            //currently only the aurora tools have a channel number that isn't 0
  for(i=0; i<numberOfTools; i++)  {
    fileNameStream.str("");
    fileNameStream<<"port_"<< this->trackerTools[i]->GetPort() + 1 <<"_"<< this->trackerTools[i]->GetChannel()<<"."<<this->currentFileNumber<<".txt";
    this->outputFiles[i].first = fileNameStream.str();
    this->outputFiles[i].second.open(this->outputFiles[i].first.c_str());
    if(this->outputFiles[i].second.fail())
      throw DataAcquisitionException(std::string("Failed to open file \"") + this->outputFiles[i].first + std::string("\" for writing."));
                         //write all digits in fixed format
    this->outputFiles[i].second<<setiosflags(ios::fixed);
  }

               //when sampleNumber is equal to zero, acquisition should run in an infinite loop (increment is zero)                 
  if(this->sampleNumber == 0) {
    currentSampleIncrement = 0;
    currentSampleSize = -1;
                      //initialize number of samples acquired per tool
    for(i=0; i<numberOfTools; i++)
      this->numAcquiredSamples[i] = -1;
  }
  else {
    currentSampleIncrement = 1;
    currentSampleSize = 0;
                      //initialize number of samples acquired per tool
    for(i=0; i<numberOfTools; i++)
      this->numAcquiredSamples[i] = -1;
  }
  while(currentSampleSize<(int)(this->sampleSize) && !this->stopAcquisition) {
    this->tracker->RequestUpdateStatus();
                           //an internal tracker error occured, couldn't get the tracking information      
    if(this->errorObserver->Error()) {
      this->errorObserver->GetErrorMessage(errorMessage);
      this->errorObserver->ClearError();
      throw DataAcquisitionException(errorMessage);        
    }
                   //the following line is part of the hack that allows us to check if a transformation is valid
                   //based on its time stamps
    currentTime = igstk::RealTimeClock::GetTimeStamp();
    displayStream.str("");
                  //get the transformation data for each of the tools
    for(j=0; j<numberOfTools; j++) {
      if(this->numAcquiredSamples[j]<this->sampleNumber) {
        displayStream<<this->outputFiles[j].first<<":\n";
        this->tracker->GetToolTransform(this->trackerTools[j]->GetPort(), this->trackerTools[j]->GetChannel(), currentTransform);        
                     //for the transformation to be valid for our purposes it must be valid as defined by its internal time span, in addition IGSTK
                     //will not update the returned transformation if the tool is not detected and it will not generate an error event. 
                     //This can result in the same transformation reported multiple times while its time span is still valid.
                     //As we don't want these untracked transforms we check that the time span is both valid with regard to the current
                     //time and that it is not equivalent to the previously reported time span.
        if(currentTransform.IsValidAtTime(currentTime) && 
           currentTransform.GetStartTime()!= prevTimeStamps[j]) {
          prevTimeStamps[j] = currentTransform.GetStartTime();          
          this->numAcquiredSamples[j]+=currentSampleIncrement;
          currentSampleSize+=currentSampleIncrement;
          igstk::Transform::VectorType translation = currentTransform.GetTranslation();
          igstk::Transform::VersorType rotation = currentTransform.GetRotation();
                                     //the time stamp is in milliseconds and has more digits beyond the decimal point which we will ignore
          this->outputFiles[j].second<<static_cast<unsigned long>(currentTransform.GetStartTime()+0.5)<<" "<<translation[0]<<" "<<translation[1]<<" "<<translation[2]<<" ";
          this->outputFiles[j].second<<rotation.GetX()<<" "<<rotation.GetY()<<" "<<rotation.GetZ()<<" "<<rotation.GetW()<<"\n";
          displayStream<<"\t"<<translation[0]<<" "<<translation[1]<<" "<<translation[2]<<" ";
          displayStream<<"\t"<<rotation.GetX()<<" "<<rotation.GetY()<<" "<<rotation.GetZ()<<" "<<rotation.GetW()<<"\n";
        }
        else
          displayStream<<"\tMISSING (not logged to file)\n";
      }
    }
                 //nofity all observers that data has been acquired
    if(this->sampleNumber == 0)
      this->InvokeEvent(DataAcquiredEvent(displayStream.str(),0));
    else
      this->InvokeEvent(DataAcquiredEvent(displayStream.str(),static_cast<double>(currentSampleSize)/static_cast<double>(sampleSize-1)));
    i++;
  }
          //close all the output streams
  for(i=0; i<numberOfTools; i++)
    this->outputFiles[i].second.close();
  delete [] prevTimeStamps;

            //nofity all observers that data acquisition has ended
  this->InvokeEvent(DataAcquisitionEndEvent());
  this->currentFileNumber++;
}
/******************************************************************************/
void DataAcquisition::ReacquireData() throw (DataAcquisitionException)
{
  this->currentFileNumber--;
  AcquireData();
}
/******************************************************************************/
void DataAcquisition::StopAcquisition()
{
  this->stopAcquisition = true;
}
