#include "DataAcquisitionSettings.h"
#include <igstkPolarisTracker.h>
#include <igstkAuroraTracker.h>
#include <sstream>


const unsigned int AuroraToolSettings::MIN_PORT_NUMBER = 0;
const unsigned int AuroraToolSettings::MAX_PORT_NUMBER = igstk::AuroraTracker::NumberOfPorts-1;

                                
const unsigned int PolarisVicraToolSettings::MIN_PORT_NUMBER = 3; //ports [0,1,2] are reserved for active tools
const unsigned int PolarisVicraToolSettings::MAX_PORT_NUMBER = igstk::PolarisTracker::NumberOfPorts-1; 

void AuroraToolSettings::SetPort(unsigned int port) throw (DataAcquisitionSettingsException)
{
  if(port<AuroraToolSettings::MIN_PORT_NUMBER || port>AuroraToolSettings::MAX_PORT_NUMBER)
    throw DataAcquisitionSettingsException("Aurora tool port is invalid, out of range of supported ports.");
  this->port = port;
}
/*****************************************************************************/
void AuroraToolSettings::SetChannel(unsigned int channel) throw (DataAcquisitionSettingsException)
{
  if(channel!=0 && channel!=1)
    throw DataAcquisitionSettingsException("Aurora tool channel is invalid (not 0 or 1).");
  this->channel = channel;
}
/*****************************************************************************/
void PolarisVicraToolSettings::SetPort(unsigned int port) throw (DataAcquisitionSettingsException)
{
  if(port<PolarisVicraToolSettings::MIN_PORT_NUMBER || port>PolarisVicraToolSettings::MAX_PORT_NUMBER) {
    std::ostringstream errorMessageStream;
    errorMessageStream<<"Polaris Vicra tool port is invalid, out of range of supported ports.";
    throw DataAcquisitionSettingsException(errorMessageStream.str());
  }
  this->port = port;
}
/*****************************************************************************/
void SerialCommunicationSettings::SetCOMPort(const unsigned int &portNumber) throw (DataAcquisitionSettingsException)
{
              //this code assumes that the igstk port numbers have a unique conversion to an integral data type
              //that should match the given port number
  switch(portNumber) {
    case igstk::SerialCommunication::PortNumber0:
    case igstk::SerialCommunication::PortNumber1:
    case igstk::SerialCommunication::PortNumber2:
    case igstk::SerialCommunication::PortNumber3:
    case igstk::SerialCommunication::PortNumber4:
    case igstk::SerialCommunication::PortNumber5:
    case igstk::SerialCommunication::PortNumber6:
    case igstk::SerialCommunication::PortNumber7:
      this->portNumber = static_cast<igstk::SerialCommunication::PortNumberType>(portNumber);
      break;
    default:
      throw DataAcquisitionSettingsException("COM port number not supported by IGSTK");
  }
}
/*****************************************************************************/

void SerialCommunicationSettings::SetBaudRate(const unsigned int & baudRate) throw (DataAcquisitionSettingsException)
{
              //this code assumes that the igstk baud rates have a unique conversion to an integral data type
              //that should match the given baud rate
  switch(baudRate) {
    case igstk::SerialCommunication::BaudRate9600 :
    case igstk::SerialCommunication::BaudRate57600 :
    case igstk::SerialCommunication::BaudRate38400 :
    case igstk::SerialCommunication::BaudRate19200 :
    case igstk::SerialCommunication::BaudRate115200 :
      this->baudRate = static_cast<igstk::SerialCommunication::BaudRateType>(baudRate);
      break;
    default:
      throw DataAcquisitionSettingsException("Invalid baud rate for serial communication.");
  }
}
/*****************************************************************************/
void SerialCommunicationSettings::SetDataBits(const unsigned int &dataBits) throw (DataAcquisitionSettingsException)
{
              //this code assumes that the igstk data bits have a unique conversion to an integral data type
              //that should match the given data bits value
  switch(dataBits) {
    case igstk::SerialCommunication::DataBits7 :
    case igstk::SerialCommunication::DataBits8 :
      this->dataBits = static_cast<igstk::SerialCommunication::DataBitsType>(dataBits);
      break;
    default:
      throw DataAcquisitionSettingsException("Invalid data bits value for serial communication.");
  }      
}
/*****************************************************************************/
void SerialCommunicationSettings::SetParity(const char &parity) throw (DataAcquisitionSettingsException)
{
              //this code assumes that the igstk parity objects have a unique conversion to an integral data type (char)
              //that should match the given parity ('N','O','E')
  switch(parity) {
    case igstk::SerialCommunication::NoParity :
    case igstk::SerialCommunication::EvenParity :
    case igstk::SerialCommunication::OddParity :
      this->parity = static_cast<igstk::SerialCommunication::ParityType>(parity);
      break;
    default:
      throw DataAcquisitionSettingsException("Invalid parity given for serial communication.");
  }      
}
/*****************************************************************************/
void SerialCommunicationSettings::SetStopBits(const unsigned int &stopBits) throw (DataAcquisitionSettingsException)
{
              //this code assumes that the igstk stop bits have a unique conversion to an integral data type
              //that should match the given stop bits value
  switch(stopBits) {
    case igstk::SerialCommunication::StopBits1 :
    case igstk::SerialCommunication::StopBits2 :
      this->stopBits = static_cast<igstk::SerialCommunication::StopBitsType>(stopBits);
      break;
    default:
      throw DataAcquisitionSettingsException("Invalid stop bits given for serial communication.");
  }      
}
/*****************************************************************************/
void SerialCommunicationSettings::SetHandshake(const bool &handshake)
{
  if(handshake)
    this->handshake = igstk::SerialCommunication::HandshakeOn;
  else 
    this->handshake = igstk::SerialCommunication::HandshakeOff;
}
/*****************************************************************************/
DataAcquisitionSettings::DataAcquisitionSettings(const DataAcquisitionSettings &other) : trackingSystemType(other.trackingSystemType),
                                                                                         serialCommunicationSettings(other.serialCommunicationSettings),
                                                                                         numSamples(other.numSamples),
                                                                                         initialFileNumber(other.initialFileNumber)
{
               //copy the tool settings data
  std::vector<NDIToolSettings *>::const_iterator it, end;
  it = other.toolSettings.begin();
  end = other.toolSettings.end();
  switch(this->trackingSystemType) {
      case TrackingSystemsEnum::POLARIS_VICRA_VALUE:
        for(; it!=end; it++) 
          this->toolSettings.push_back(new PolarisVicraToolSettings(*(dynamic_cast<PolarisVicraToolSettings *>(*it))));
        break;
      case TrackingSystemsEnum::AURORA_VALUE:
        for(; it!=end; it++) 
          this->toolSettings.push_back(new AuroraToolSettings(*(dynamic_cast<AuroraToolSettings *>(*it))));
        break;
  }
}
/*****************************************************************************/
DataAcquisitionSettings::~DataAcquisitionSettings() 
{
  std::vector<NDIToolSettings *>::iterator it, end;
  it = this->toolSettings.begin();
  end = this->toolSettings.end();
  for(; it!=end; it++) 
    delete (*it);
}
/*****************************************************************************/
DataAcquisitionSettings & DataAcquisitionSettings::operator=(DataAcquisitionSettings &right) 
{
  this->trackingSystemType = right.trackingSystemType;
  this->serialCommunicationSettings = right.serialCommunicationSettings;
  this->numSamples = right.numSamples;
  this->initialFileNumber = right.initialFileNumber;
              //copy the tool settings
  std::vector<NDIToolSettings *>::iterator it, end;
  it = right.toolSettings.begin();
  end = right.toolSettings.end();
  switch(this->trackingSystemType) {
      case TrackingSystemsEnum::POLARIS_VICRA_VALUE:
        for(; it!=end; it++) 
          this->toolSettings.push_back(new PolarisVicraToolSettings(*(dynamic_cast<PolarisVicraToolSettings *>(*it))));
        break;
      case TrackingSystemsEnum::AURORA_VALUE:
        for(; it!=end; it++) 
          this->toolSettings.push_back(new AuroraToolSettings(*(dynamic_cast<AuroraToolSettings *>(*it))));
        break;
  }
  return *this;
}
/*****************************************************************************/
void DataAcquisitionSettings::SetToolSettings(const std::vector<NDIToolSettings *> &toolSettings) 
{
  std::vector<NDIToolSettings *>::const_iterator it, end;
                //clear the current tool settings
  it = this->toolSettings.begin();
  end = this->toolSettings.end();
  for(; it!=end; it++) 
    delete (*it);
  this->toolSettings.clear();

                  //copy the new tool settings
  it = toolSettings.begin();
  end = toolSettings.end();
  switch(this->trackingSystemType) {
      case TrackingSystemsEnum::POLARIS_VICRA_VALUE:
        for(; it!=end; it++) 
          this->toolSettings.push_back(new PolarisVicraToolSettings(*(dynamic_cast<PolarisVicraToolSettings *>(*it))));
        break;
      case TrackingSystemsEnum::AURORA_VALUE:
        for(; it!=end; it++) 
          this->toolSettings.push_back(new AuroraToolSettings(*(dynamic_cast<AuroraToolSettings *>(*it))));
        break;
  }
}
/*****************************************************************************/
void DataAcquisitionSettings::GetToolSettings(std::vector<NDIToolSettings *> &toolSettings) 
{
  std::vector<NDIToolSettings *>::const_iterator it, end;
  it = this->toolSettings.begin();
  end = this->toolSettings.end();
  switch(this->trackingSystemType) {
      case TrackingSystemsEnum::POLARIS_VICRA_VALUE:
        for(; it!=end; it++) 
          toolSettings.push_back(new PolarisVicraToolSettings(*(dynamic_cast<PolarisVicraToolSettings *>(*it))));
        break;
      case TrackingSystemsEnum::AURORA_VALUE:
        for(; it!=end; it++) 
          toolSettings.push_back(new AuroraToolSettings(*(dynamic_cast<AuroraToolSettings *>(*it))));
        break;
  }  
}
