#ifndef _DATA_ACQUISITION_SETTINGS_H_
#define _DATA_ACQUISITION_SETTINGS_H_

/**************************************************************************
* Copyright (c) 2007
* Computer Aided Interventions and Medical Robotics (CAIMR) group,
* Imaging Science and Information Systems (ISIS) research center,
* Georgetown University.
*
* All rights reserved.
*
* Permission to use, copy, modify, distribute, and sell this software
* and its documentation for any purpose is hereby granted without fee,
* provided that:
* (1) Redistributions of source code must retain the above copyright notice,
*     this list of conditions and the following disclaimer.
* (2) Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions and the following disclaimer in the documentation
*      and/or other materials provided with the distribution.
* (3) The author's name may not be used to endorse or promote products derived
*     from this software without specific prior written permission.
* (4) Modified source versions must be plainly marked as such, and must not be
*     misrepresented as being the original software.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ``AS IS'' AND ANY EXPRESS
* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
* OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
* WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
* OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/


#include <igstkSerialCommunication.h>
#include "TrackingSystemsEnum.h"

/**
 * The exception type thrown by the methods associated with the DataAcquisitionSettings class and its component classes.
 *
 * @author Ziv Yaniv
 */
class DataAcquisitionSettingsException : public std::exception
{
public:
  /**
   * Construct an exception with a specifc message.
   */
  DataAcquisitionSettingsException(const std::string &errorMessage) {this->errorMessage = errorMessage;}

  /**
   * Virtual destructor.
   */
  virtual ~DataAcquisitionSettingsException() throw() {}

  /**
   * Get the error message.
   */
   virtual const char* what() const throw() {return errorMessage.c_str();}
private:
  std::string errorMessage;
};


/**
 * Each tool used with the NDI trackers has a port, a channel, and a tool description.  
 * If the tool does not have its description on its SROM it is loaded
 * from file (this is always the case with passive tools and some active ones).
 * An empty SROM file name indicates that this tool already has its description on
 * its SROM.
 */
class NDIToolSettings {
public:
  /**
   * Construct a new NDI tool settings object. The default port and channel are set to 0, and no SROM file is used.
   */
  NDIToolSettings() : port(0), channel(0) {}

  /**
   * Copy constructor for NDI tool settings object. Just copies the port and srom file name from the
   * other tool.
   */
  NDIToolSettings(const NDIToolSettings &other) : port(other.port), channel(other.channel), sromFileName(other.sromFileName) {}
  
  /**
   * virtual destructor.
   */
  virtual ~NDIToolSettings(){}

  /**
   * Set the port for this tool. Possible port numbers are dependent upon the specific tracking system.
   */
  virtual void SetPort(unsigned int port) throw (DataAcquisitionSettingsException) = 0;

  /**
   * Get the tool's port number.
   */
  unsigned int GetPort() {return this->port;}

  /**
   * Get the tool's channel number.
   */
  unsigned int GetChannel() {return this->channel;}

  /**
   * Set the name of the SROM file which is loaded into the tool's SROM. An empty SROM
   * file name indicates that the tool's description is already on its SROM and does not
   * need to be loaded from file.
   */
  void SetSROMFileName(std::string &sromFileName) {this->sromFileName = sromFileName;}
  void SetSROMFileName(std::string sromFileName) {this->sromFileName = sromFileName;}

  /**
   * Get the name of the tool's SROM file. An empty string denotes no SROM file.
   */
  const std::string & GetSROMFileName() {return sromFileName;}
protected:
  unsigned int port;
  unsigned int channel;
  std::string sromFileName;
};

/**
 * Aurora tools have both a port and a channel. A port relates to the physical port on the aurora control
 * box. A single port can be used with a splitter to track two five degree of freedom tools, providing
 * two data channels on one port. Thus, to uniquely specify a tool both its port and
 * channel are required.
 */
class AuroraToolSettings : public NDIToolSettings {
public:
  /**
   * Construct a new aurora tool settings object. The default port is 0, the default channel is 0, and no SROM file is used.
   */
  AuroraToolSettings() : NDIToolSettings() {}

  AuroraToolSettings(const AuroraToolSettings &other) : NDIToolSettings(other) {}
  
  virtual ~AuroraToolSettings(){}
  
  AuroraToolSettings & operator=(const AuroraToolSettings &right) {this->port = right.port; this->channel = right.channel; this->sromFileName = right.sromFileName; return *this;}
  
  /**
   * Set the tool's port number. 
   * @param port The port number to set. Valid values are in [AuroraToolSettings::MIN_PORT_NUMBER,AuroraToolSettings::MAX_PORT_NUMBER].
   */
  virtual void SetPort(unsigned int port) throw (DataAcquisitionSettingsException);

  /**
   * Set the tool's channel number.
   * @param channel The channel to set. Valid values are 0,1.
   */
  void SetChannel(unsigned int channel) throw (DataAcquisitionSettingsException);

private:
    static const unsigned int MIN_PORT_NUMBER;
    static const unsigned int MAX_PORT_NUMBER;
};  

/**
 * Polaris vicra tools are uniquely idenfitied by their port. 
 */
class PolarisVicraToolSettings : public NDIToolSettings {
public:
  PolarisVicraToolSettings() : NDIToolSettings() {}

  PolarisVicraToolSettings(const PolarisVicraToolSettings &other) : NDIToolSettings(other) {}

  virtual ~PolarisVicraToolSettings(){}

                                                                           //no need to copy channel as this is zero for all polaris tools
  PolarisVicraToolSettings & operator=(const PolarisVicraToolSettings &right) {this->port = right.port; this->sromFileName = right.sromFileName; return *this;}
  
  /**
   * Set the tool's port number. Valid numbers are in [PolarisVicraToolSettings::MIN_PORT_NUMBER,PolarisVicraToolSettings::MAX_PORT_NUMBER].
   */
  virtual void SetPort(unsigned int port) throw (DataAcquisitionSettingsException);
private:
    static const unsigned int MIN_PORT_NUMBER;
    static const unsigned int MAX_PORT_NUMBER;
};  


/**
 * Settings for the serial communication.
 */
class SerialCommunicationSettings {
public:
  SerialCommunicationSettings() : portNumber(igstk::SerialCommunication::PortNumber0),
                                  baudRate(igstk::SerialCommunication::BaudRate115200),
                                  dataBits(igstk::SerialCommunication::DataBits8),
                                  parity(igstk::SerialCommunication::NoParity),
                                  stopBits(igstk::SerialCommunication::StopBits1),
                                  handshake(igstk::SerialCommunication::HandshakeOff) {}

  SerialCommunicationSettings(const SerialCommunicationSettings &other) : portNumber(other.portNumber),
                                                                          baudRate(other.baudRate),
                                                                          dataBits(other.dataBits),
                                                                          parity(other.parity),
                                                                          stopBits(other.stopBits),
                                                                          handshake(other.handshake) {}
  virtual ~SerialCommunicationSettings(){}

  SerialCommunicationSettings & operator=(const SerialCommunicationSettings &right) {
    this->portNumber = right.portNumber;
    this->baudRate = right.baudRate;
    this->dataBits = right.dataBits;
    this->parity = right.parity;
    this->stopBits = right.stopBits;
    this->handshake = right.handshake;
    return *this;
  }
  
  /**
   * Set the COM port number.
   * @param portNumber The COM port number through which we communicate with the tracking system. See 
   *                   igstk::SerialCommunication for valid values.
   */
  void SetCOMPort(const unsigned int &portNumber) throw (DataAcquisitionSettingsException);
  igstk::SerialCommunication::PortNumberType GetCOMPort() {return this->portNumber;}

  /**
   * Set the baud rate.
   * @param baudRate The baud rate at which we communicate with the tracking system. See 
   *                   igstk::SerialCommunication for valid values.
   */
  void SetBaudRate(const unsigned int & baudRate) throw (DataAcquisitionSettingsException);
  igstk::SerialCommunication::BaudRateType GetBaudRate() {return this->baudRate;}

  /**
   * Set the data bits value.
   * @param dataBits The data bits value for communicating with the tracking system. See 
   *                   igstk::SerialCommunication for valid values.
   */
  void SetDataBits(const unsigned int &dataBits) throw (DataAcquisitionSettingsException);
  igstk::SerialCommunication::DataBitsType GetDataBits() {return this->dataBits;}

  /**
   * Set the parity value.
   * @param parity The parity value for communicating with the tracking system. See 
   *                   igstk::SerialCommunication for valid values.
   */
  void SetParity(const char &parity) throw (DataAcquisitionSettingsException);
  igstk::SerialCommunication::ParityType GetParity() {return this->parity;}

  /**
   * Set the stop bits value.
   * @param stopBits The stop bits value for communicating with the tracking system. See 
   *                   igstk::SerialCommunication for valid values.
   */
  void SetStopBits(const unsigned int &stopBits) throw (DataAcquisitionSettingsException);
  igstk::SerialCommunication::StopBitsType GetStopBits() {return this->stopBits;}

  /**
   * Set the hand shake value.
   * @param handshake The hand shake value for communicating with the tracking system. See 
   *                   igstk::SerialCommunication for valid values.
   */
  void SetHandshake(const bool &handshake);
  igstk::SerialCommunication::HandshakeType GetHandshake() {return this->handshake;}
private:
          //all the settings needed for serial communication
  igstk::SerialCommunication::PortNumberType portNumber;
  igstk::SerialCommunication::BaudRateType baudRate;
  igstk::SerialCommunication::DataBitsType dataBits;
  igstk::SerialCommunication::ParityType parity;
  igstk::SerialCommunication::StopBitsType stopBits;
  igstk::SerialCommunication::HandshakeType handshake;
};


/**
 * The settings for data acquisition. These include the tracking system type, the communication settings, the number of
 * samples per tool for each acquisition run, and the tool settings.
 * When the number of samples is set to zero the data acquisition will continue until the user indicates that it stop.
 */
class DataAcquisitionSettings {
public:
  /**
   * Default constructor. Tracking system is set to polaris with default communication settings,
   * continueous data logging (numSamples==0), initial file number is 0, and no tools.
   */
  DataAcquisitionSettings() : trackingSystemType(TrackingSystemsEnum::POLARIS_VICRA), numSamples(0), initialFileNumber(0) {}

  DataAcquisitionSettings(const DataAcquisitionSettings &other);
  
  virtual ~DataAcquisitionSettings();

  DataAcquisitionSettings & operator=(DataAcquisitionSettings &right);

  void SetTrackingSytemType(const TrackingSystemsEnum &trackingSystemType) {this->trackingSystemType = trackingSystemType;}
  TrackingSystemsEnum GetTrackingSystemType() {return this->trackingSystemType;}

  void SetSerialCommunicationSettings(const SerialCommunicationSettings &serialCommunicationSettings) {this->serialCommunicationSettings = serialCommunicationSettings;}
  SerialCommunicationSettings GetSerialCommunicationSettings() {return this->serialCommunicationSettings;}

  void SetToolSettings(const std::vector<NDIToolSettings *> &toolSettings); 

  /**
   * The tool settings are copied into the given vector. It is the callers responsiblity to release the
   * memory.
   */
  void GetToolSettings(std::vector<NDIToolSettings *> &toolSettings);

  void SetSampleNumber(unsigned int numSamples) {this->numSamples = numSamples;}
  unsigned int GetSampleNumber() {return this->numSamples;}

  void SetInitialFileNumber(unsigned int initialFileNumber) {this->initialFileNumber = initialFileNumber;}
  unsigned int GetInitialFileNumber() {return initialFileNumber;}

  
private:
              //which tracking system are we using
  TrackingSystemsEnum trackingSystemType;
              //serial communication settings used
  SerialCommunicationSettings serialCommunicationSettings;
        //information on the tools whose tracking data we want to record
  std::vector<NDIToolSettings *> toolSettings;  
       //number of samples acquired for each tool
  unsigned int numSamples;
              //initial file number used for logging the tracking data
  unsigned int initialFileNumber;
};

#endif _DATA_ACQUISITION_SETTINGS_H_
