#include <itksys/SystemTools.hxx>
#include <itkExceptionObject.h>

#include "DataAcquisitionSettingsFileReader.h"
#include "GeneralUtilityMethods.h"

DataAcquisitionSettingsFileReader::DataAcquisitionSettingsFileReader (const std::string *settingsFileName) : readSettings(false),
                                                                                                             fileNameSet(false)
{
  if(settingsFileName != NULL) {
    this->settingsFileName = *settingsFileName;
    this->fileNameSet = true;
  }
}
/*****************************************************************************/
void DataAcquisitionSettingsFileReader::SetFileName(const std::string &settingsFileName)
{
  this->settingsFileName = settingsFileName;
  this->fileNameSet = true;
  this->readSettings = false;
}
/*****************************************************************************/
const std::string & DataAcquisitionSettingsFileReader::GetFileName() throw (DataAcquisitionSettingsFileReaderException)
{
  if(!this->fileNameSet)
    throw DataAcquisitionSettingsFileReaderException(std::string("Cannot get file name as it was not set."));

  return this->settingsFileName;
}
/*****************************************************************************/
void DataAcquisitionSettingsFileReader::Read() throw (DataAcquisitionSettingsFileReaderException)
{
  if(!this->fileNameSet)
    throw DataAcquisitionSettingsFileReaderException(std::string("Cannot read settings file, as file name was not set."));

                    //create an instance of the class that actually does the parsing
  XMLDataAcquisitionSettingsFileReader::Pointer xmlReader = XMLDataAcquisitionSettingsFileReader::New();
  xmlReader->SetSettings(&this->settings);
                     //check that we can actually read the file
  if(xmlReader->CanReadFile(this->settingsFileName.c_str())) {
                     //set the file name
    xmlReader->SetFilename(this->settingsFileName.c_str());
    try {       //read the xml
      xmlReader->GenerateOutputInformation();
      this->readSettings = xmlReader->ReadSettings();
    } catch(itk::ExceptionObject &itkEO) { //an itk exception is thrown if the xml is malformed, for uniformity I
                                          //catch it and throw my own exception
      throw DataAcquisitionSettingsFileReaderException(std::string(itkEO.GetDescription()));
    }
  }
}
/*****************************************************************************/
void DataAcquisitionSettingsFileReader::GetSettings(DataAcquisitionSettings &settings) throw (DataAcquisitionSettingsFileReaderException)
{
  if(!this->readSettings)
    throw DataAcquisitionSettingsFileReaderException("Cannot get settings, as configuration file has not been read.");
  settings = this->settings;  
}
/*****************************************************************************/
int DataAcquisitionSettingsFileReader::XMLDataAcquisitionSettingsFileReader::CanReadFile(const char* name)
{
       //check that we have a valid file name, it is not a directory, and the file exists
  if(itksys::SystemTools::FileLength(name) == 0 ||
     !itksys::SystemTools::FileExists(name) ||
     itksys::SystemTools::FileIsDirectory(name))
      return 0;
  return 1;
}
/*****************************************************************************/
/*
 * The attributes come in pairs of name and value. If the xml is malformed, incomplete
 * attributes this method will not be invoked.
 */
void DataAcquisitionSettingsFileReader::XMLDataAcquisitionSettingsFileReader::StartElement(const char * name,const char **atts)
{
            //we have a acquisition_configuration section       
  if(itksys::SystemTools::Strucmp(name,"acquisition_configuration") == 0) {
                   //there can only be one "acquisition_configuration" per xml file                  
    if(this->readSettings)
      throw DataAcquisitionSettingsFileReaderException(std::string("The tag \"acquisition_configuration\" appears multiple times, should only appear once."));
    ProcessDataAcquisitionSettings(atts);
    this->readingAcquisitionSettings = true;
  }   
                //we have a tracking system settings section
  if(this->readingAcquisitionSettings && itksys::SystemTools::Strucmp(name,"tracking_system") == 0) {
    ProcessTrackingSystemSettings(atts);
    this->readingTrackerSettings = true;
  }

                 //we have the communication section settings for the DataAcquisition system  
  if(this->readingTrackerSettings && itksys::SystemTools::Strucmp(name,"communication") == 0) 
    ProcessCommunicationSettings(atts);

            //we have a tool section settings for the DataAcquisition system
  if(this->readingTrackerSettings && itksys::SystemTools::Strucmp(name,"tool") == 0) {
    switch(this->settings->GetTrackingSystemType()) {
      case TrackingSystemsEnum::POLARIS_VICRA_VALUE:
        ProcessPolarisToolSettings(atts);
        break;
      case TrackingSystemsEnum::AURORA_VALUE:
        ProcessAuroraToolSettings(atts);
        break;
    }    
  }
}
/*****************************************************************************/
void DataAcquisitionSettingsFileReader::XMLDataAcquisitionSettingsFileReader::ProcessDataAcquisitionSettings(const char **atts) throw (DataAcquisitionSettingsFileReaderException)
{
  bool haveSamples, haveInitialFileNumber;
  haveSamples = haveInitialFileNumber = false;
                
  for(int i=0; atts[i] != NULL; i+=2) {
            //deal with the "samples" attribute
    if(itksys::SystemTools::Strucmp(atts[i],"samples") == 0) {
      int samples;
      try {
        samples = GeneralUtilityMethods::Atoi(atts[i+1]);
      } 
      catch(std::runtime_error &) {
        throw DataAcquisitionSettingsFileReaderException("Non-integral samples number given for samples attribute.");
      }
      if(samples<0)
        throw DataAcquisitionSettingsFileReaderException("Negative number given as number of samples attribute.");
      haveSamples = true;
      this->settings->SetSampleNumber(static_cast<unsigned int>(samples));
    }       //deal with the "initial_file_number" attribute       
    if(itksys::SystemTools::Strucmp(atts[i],"initial_file_number") == 0) {
      int initialFileNumber;
      try {
        initialFileNumber = GeneralUtilityMethods::Atoi(atts[i+1]);
      }  
      catch(std::runtime_error &) {
        throw DataAcquisitionSettingsFileReaderException("Non-integral initial file number given for \"initial_file_number\" attribute.");
      }
      haveInitialFileNumber = true;
      this->settings->SetInitialFileNumber(static_cast<unsigned int>(initialFileNumber));
    }
  }
}
/*****************************************************************************/
void DataAcquisitionSettingsFileReader::XMLDataAcquisitionSettingsFileReader::ProcessTrackingSystemSettings(const char **atts) throw (DataAcquisitionSettingsFileReaderException)
{
  bool haveTrackingSystemType = false;

  for(int i=0; atts[i] != NULL; i+=2) {
    if(itksys::SystemTools::Strucmp(atts[i],"type") == 0) {
      if(itksys::SystemTools::Strucmp(atts[i+1],TrackingSystemsEnum::POLARIS_VICRA.ToString().c_str()) == 0) {
        this->settings->SetTrackingSytemType(TrackingSystemsEnum::POLARIS_VICRA);
        haveTrackingSystemType = true;
      }
      if(itksys::SystemTools::Strucmp(atts[i+1],TrackingSystemsEnum::AURORA.ToString().c_str()) == 0) {
        this->settings->SetTrackingSytemType(TrackingSystemsEnum::AURORA);
        haveTrackingSystemType = true;
      }
    }
  }
  if(!haveTrackingSystemType)
    throw DataAcquisitionSettingsFileReaderException(std::string("Unsupported or missing \"type\" attribute in tracking_system settings."));  
}
/*****************************************************************************/
void DataAcquisitionSettingsFileReader::XMLDataAcquisitionSettingsFileReader::ProcessCommunicationSettings(const char **atts) throw (DataAcquisitionSettingsFileReaderException)
{
  int i;
  bool haveComPort, haveBaudRate, haveDataBits, haveParity, haveStopBits, haveHandshake;
  haveComPort = haveBaudRate = haveDataBits = haveParity = haveStopBits = haveHandshake = false;
  SerialCommunicationSettings communicationSettings;

  for(i=0; atts[i] != NULL; i+=2) {
                          //get com port number attribute
    if(itksys::SystemTools::Strucmp(atts[i],"com_port") == 0) {
      unsigned int portNumber;
      try {
        portNumber = GeneralUtilityMethods::Atoi(atts[i+1]);
            //in the xml file we have the com number, but the internal igstk constants start from zero
        portNumber--;
        communicationSettings.SetCOMPort(portNumber);
        haveComPort = true;
      }
      catch(DataAcquisitionSettingsException &dase) { //COM port not supported by IGSTK
        throw DataAcquisitionSettingsFileReaderException(dase.what());
      }
      catch(std::runtime_error &) {
        throw DataAcquisitionSettingsFileReaderException("Illegal port number format given for DataAcquisition system, expected unsigned int.");
      }
    }
                           //get baud rate attribute
    if(itksys::SystemTools::Strucmp(atts[i],"baud_rate") == 0) {
      unsigned int baudRate;
      try {
        baudRate = GeneralUtilityMethods::Atoi(atts[i+1]);
        communicationSettings.SetBaudRate(baudRate);
        haveBaudRate = true;
      }
      catch(DataAcquisitionSettingsException &dase) { //baud rate not supported by IGSTK
        throw DataAcquisitionSettingsFileReaderException(dase.what());
      }
      catch(std::runtime_error &) {
        throw DataAcquisitionSettingsFileReaderException("Illegal baud rate format for DataAcquisition system, expected unsigned int.");
      }
    }

                             //get data bits attribute
    if(itksys::SystemTools::Strucmp(atts[i],"data_bits") == 0) {
      unsigned int dataBits;
      try {
        dataBits = GeneralUtilityMethods::Atoi(atts[i+1]);
        communicationSettings.SetDataBits(dataBits);
        haveDataBits = true;
      }
      catch(DataAcquisitionSettingsException &dase) { //invalid data bits value 
        throw DataAcquisitionSettingsFileReaderException(dase.what());
      }
      catch(std::runtime_error &) {
        throw DataAcquisitionSettingsFileReaderException("Illegal data bits format for DataAcquisition system, expected be unsigned int.");
      }
    }
             //get the parity type attribute
    if(itksys::SystemTools::Strucmp(atts[i],"parity") == 0) {
        char parity = atts[i+1][0];
        try {
          communicationSettings.SetParity(parity);
          haveParity = true;
        }
        catch(DataAcquisitionSettingsException &dase) { //invalid parity value 
          throw DataAcquisitionSettingsFileReaderException(dase.what());
        }
    }
           //get the stop bits attribute
    if(itksys::SystemTools::Strucmp(atts[i],"stop_bits") == 0) {
      unsigned int stopBits;
      try {
        stopBits = GeneralUtilityMethods::Atoi(atts[i+1]);
        communicationSettings.SetStopBits(stopBits);
        haveStopBits = true;
      }
      catch(DataAcquisitionSettingsException &dase) { //invalid stop bits value 
        throw DataAcquisitionSettingsFileReaderException(dase.what());
      }
      catch(std::runtime_error &) {
        throw DataAcquisitionSettingsFileReaderException("Illegal stop bits format for DataAcquisition system, expected unsigned int.");
      }
    }
                        //get the hand shake attribute
    if(itksys::SystemTools::Strucmp(atts[i],"hand_shake") == 0) {
      bool handshake;
      try {
        handshake = (GeneralUtilityMethods::Atoi(atts[i+1]) == 1);
        communicationSettings.SetHandshake(handshake);
        haveHandshake = true;
      }
      catch(std::runtime_error &) {
        throw DataAcquisitionSettingsFileReaderException("Illegal hand_shake format for DataAcquisition system, should be 0 or 1.");
      }
    }
  }
  if(! (haveComPort && haveBaudRate && haveDataBits && haveParity && haveStopBits && haveHandshake))
    throw DataAcquisitionSettingsFileReaderException(std::string("Missing attribute in communication settings."));
  this->settings->SetSerialCommunicationSettings(communicationSettings);
}
/*****************************************************************************/
/**
 * A polaris tool has two attributes, port, and SROM file name. The port
 * is the only required attribute.
 */
void DataAcquisitionSettingsFileReader::XMLDataAcquisitionSettingsFileReader::ProcessPolarisToolSettings(const char **atts) throw (DataAcquisitionSettingsFileReaderException)
{
  PolarisVicraToolSettings *polarisTool = new PolarisVicraToolSettings(); //by default port is 0 and no SROM file
  unsigned int portNumber;

  for(int i=0; atts[i] != NULL; i+=2) {
                //get the port number
    if(itksys::SystemTools::Strucmp(atts[i],"port") == 0) {
      try {
        portNumber = GeneralUtilityMethods::Atoi(atts[i+1]);
        portNumber-=1; //the user assigns ports starting at index 1, internally they start at index 0
        polarisTool->SetPort(portNumber);
      }
      catch(DataAcquisitionSettingsException &dase) { //invalid tool port value
        throw DataAcquisitionSettingsFileReaderException(dase.what());
      }
      catch(std::runtime_error &) {
        throw DataAcquisitionSettingsFileReaderException("Illegal tool port value given for DataAcquisition system, should be unsigned int.");
      }
    }
                 //get the SROM file name if provided
    if(itksys::SystemTools::Strucmp(atts[i],"srom_file") == 0) {
      polarisTool->SetSROMFileName(std::string(atts[i+1]));  
    }
  }
  portNumber = polarisTool->GetPort();
     //the port number has to be unique
  std::vector<NDIToolSettings *>::iterator it = this->tmpToolSettings.begin();
  std::vector<NDIToolSettings *>::iterator toolsEnd = this->tmpToolSettings.end();
            //the loop assumes short circuit evaluation
  for(; it!=toolsEnd && (*it)->GetPort()!=portNumber; it++);
  if(it!=toolsEnd) {
    for(it = this->tmpToolSettings.begin(); it!=toolsEnd; it++)
      delete (*it);
    throw DataAcquisitionSettingsFileReaderException(std::string("Multiple tools assigned the same port."));
  }
  this->tmpToolSettings.push_back(polarisTool);
}
/*****************************************************************************/
/**
 * An aurora tool has three attributes, port, channel, and SROM file name. The port
 * and channel are required attributes (when a single port is used with a splitter to
 * track two five degree of freedom sensors each of them is identified by its data 
 * channel on the port).
 */
void DataAcquisitionSettingsFileReader::XMLDataAcquisitionSettingsFileReader::ProcessAuroraToolSettings(const char **atts) throw (DataAcquisitionSettingsFileReaderException)
{
  AuroraToolSettings *auroraTool = new AuroraToolSettings(); //by default port and channel are 0 and no SROM file
  unsigned int portNumber, channelNumber;

  for(int i=0; atts[i] != NULL; i+=2) {
                //get the port number
    if(itksys::SystemTools::Strucmp(atts[i],"port") == 0) {
      try {
        portNumber = GeneralUtilityMethods::Atoi(atts[i+1]);
        portNumber-=1; //the user assigns ports starting at index 1, internally they start at index 0
        auroraTool->SetPort(portNumber);   
      }
      catch(DataAcquisitionSettingsException &dase) { //invalid tool port value
        throw DataAcquisitionSettingsFileReaderException(dase.what());
      }
      catch(std::runtime_error &) {
        throw DataAcquisitionSettingsFileReaderException("Illegal tool port value given for DataAcquisition system, should be unsigned int.");
      }
    }
                //get the channel number
    if(itksys::SystemTools::Strucmp(atts[i],"channel") == 0) {
      try {
        channelNumber = GeneralUtilityMethods::Atoi(atts[i+1]);
        auroraTool->SetChannel(channelNumber);       
      }
      catch(DataAcquisitionSettingsException &dase) { //invalid channel value
        throw DataAcquisitionSettingsFileReaderException(dase.what());
      }
      catch(std::runtime_error &) {
        throw DataAcquisitionSettingsFileReaderException("Illegal channel value given for DataAcquisition system, should be unsigned int.");
      }
    }
                 //get the SROM file name if provided
    if(itksys::SystemTools::Strucmp(atts[i],"srom_file") == 0) {
      auroraTool->SetSROMFileName(std::string(atts[i+1]));  
    }
  }
  portNumber = auroraTool->GetPort();
  channelNumber = auroraTool->GetChannel();
     //the combination of port and channel numbers has to be unique
  std::vector<NDIToolSettings *>::iterator it = this->tmpToolSettings.begin();
  std::vector<NDIToolSettings *>::iterator toolsEnd = this->tmpToolSettings.end();
  while(it!=toolsEnd) {
    AuroraToolSettings *aToolSettings = dynamic_cast<AuroraToolSettings *>(*it);
                   //non unique tool, delete all tools and throw an exception
    if(aToolSettings->GetPort() == portNumber && aToolSettings->GetChannel() == channelNumber) {
      for(it = this->tmpToolSettings.begin(); it!=toolsEnd; it++)
        delete (*it);
      throw DataAcquisitionSettingsFileReaderException(std::string("Multiple tools assigned the same port and channel."));
    }
    it++;
  }
  this->tmpToolSettings.push_back(auroraTool);
}
/*****************************************************************************/
void DataAcquisitionSettingsFileReader::XMLDataAcquisitionSettingsFileReader::EndElement(const char *name)
{
        //got the end tag for a tracking system                 
	if(this->readingTrackerSettings && itksys::SystemTools::Strucmp(name,"tracking_system") == 0) {
		this->readingTrackerSettings = false;
                          //copy all tool settings into the settings object
    this->settings->SetToolSettings(this->tmpToolSettings);
    for(std::vector<NDIToolSettings *>::iterator it = this->tmpToolSettings.begin(); it!=this->tmpToolSettings.end(); it++)
      delete (*it);
	}
        //got the end tag for the acquisition configuration      
  if(this->readingAcquisitionSettings  && itksys::SystemTools::Strucmp(name,"acquisition_configuration") == 0) {
		this->readingAcquisitionSettings = false;		
    this->readSettings = true;
	}
}
