#ifndef _DATA_ACQUISITION_SETTINGS_FILE_READER_H_
#define _DATA_ACQUISITION_SETTINGS_FILE_READER_H_

/**************************************************************************
* Copyright (c) 2007
* Computer Aided Interventions and Medical Robotics (CAIMR) group,
* Imaging Science and Information Systems (ISIS) research center,
* Georgetown University.
*
* All rights reserved.
*
* Permission to use, copy, modify, distribute, and sell this software
* and its documentation for any purpose is hereby granted without fee,
* provided that:
* (1) Redistributions of source code must retain the above copyright notice,
*     this list of conditions and the following disclaimer.
* (2) Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions and the following disclaimer in the documentation
*      and/or other materials provided with the distribution.
* (3) The author's name may not be used to endorse or promote products derived
*     from this software without specific prior written permission.
* (4) Modified source versions must be plainly marked as such, and must not be
*     misrepresented as being the original software.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ``AS IS'' AND ANY EXPRESS
* OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
* OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
* WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
* OF THE POSSIBILITY OF SUCH DAMAGE.
**************************************************************************/

#include <stdexcept>
#include <itkXMLFile.h>
#include <igstkSerialCommunication.h>

//#include "DataAcquisition.h"
#include "DataAcquisitionSettings.h"
#include "TrackingSystemsEnum.h"

/**
 * The exception type thrown by the DataAcquisitionSettingsFileReader methods.
 *
 * @author Ziv Yaniv
 */
class DataAcquisitionSettingsFileReaderException : public std::exception
{
public:
  /**
   * Construct an exception with a specifc message.
   */
  DataAcquisitionSettingsFileReaderException(const std::string &errorMessage) {this->errorMessage = errorMessage;}

  /**
   * Virtual destructor.
   */
  virtual ~DataAcquisitionSettingsFileReaderException() throw() {}

  /**
   * Get the error message.
   */
   virtual const char* what() const throw() {return errorMessage.c_str();}
private:
  std::string errorMessage;
};
/*****************************************************************************/

/**
 * A reader that reads an XML settings file containing the settings for the serial communication and desired tool
 * ports and SROM file locations.
 * 
 * This class only provides an interface for reading the XML settings file and getting the data. 
 * An internal class inherited from the itk::XMLReaderBase does the actual parsing.
 *
 * @author Ziv Yaniv
 */
class DataAcquisitionSettingsFileReader 
{
public:
  /**
   * Construct a new file reader.
   * @param settingsFileName If a non NULL string is given this is the name of the settings file.
   */
  DataAcquisitionSettingsFileReader(const std::string *settingsFileName = NULL);

  /**
   * Set the settings file name.
   */
  void SetFileName(const std::string &settingsFileName);

  /**
   * Get the current settings file name.
   * @exception DataAcquisitionSettingsFileReaderException Throws an exception if the file name was not previously set
   * either in the constructor or by the SetFileName() method.
   */
  const std::string & GetFileName() throw (DataAcquisitionSettingsFileReaderException);

  /**
   * Read the current settings file.
   * @exception ExceptionObject An exception is thrown if there is a problem
   *                            reading the file, or the file is not well formed 
   *                            (does not follow the rules for valid xml).
   * @exception DataAcquisitionSettingsFileReaderException An exception is thrown if there are
   *                                              missing settings parameters or they have illegal values.
   */
  void Read() throw (DataAcquisitionSettingsFileReaderException);

  /**
   * This method checks if settings were actually read, as reading the settings file is successful if it 
   * is a valid xml file, even if it doesn't contain settings for the data acquisition.
   * 
   */
  bool ReadSettings() {return readSettings;}

  /**
   * Get the data acquisition settings.
   * @exception DataAcquisitionSettingsFileReaderException Throws this exception if the current settings file 
   *                                                       hasn't been read.
   */
  void GetSettings(DataAcquisitionSettings &settings) throw (DataAcquisitionSettingsFileReaderException);

private:
  
	/**
	 * This is the class that actually does the XML reading and analysis.
	 * The style follows the ITK style as it is sub-classed from itk::XMLReaderBase
	 */
  class XMLDataAcquisitionSettingsFileReader : public itk::XMLReaderBase 
  {
  public:
    /** Standard typedefs */ 
    typedef XMLDataAcquisitionSettingsFileReader           Self;
    typedef itk::XMLReaderBase                             Superclass;
    typedef itk::SmartPointer<Self>                        Pointer;

           /** Run-time type information (and related methods). */
    itkTypeMacro(XMLDataAcquisitionSettingsFileReader, itk::XMLReaderBase);

            /** Method for creation through the object factory. */
    itkNewMacro(Self);

    /**
     * Set the pointer to the settings object that is filled when reading the xml file.
     * It is assumed that the pointer is not NULL. This is reasonable as 
     * XMLDataAcquisitionSettingsFileReader is an inner class of DataAcquisitionSettingsFileReader
     * and is not accessible outside of it.
     */
    void SetSettings(DataAcquisitionSettings *settings) {this->settings = settings;}

    /**
     * Check that the given name isn't a zero length string, that the file exists, and that it
     * isn't a directory.
     */
    virtual int CanReadFile(const char* name);

    /**
     * Method called when a new xml tag start is encountered.
     */
    virtual void StartElement(const char * name,const char **atts); 

    /**
     * Method called when an xml tag end is encountered.
     */
    virtual void EndElement(const char *name); 

    /**
     * Method for handling the data inside an xml tag.
     * I opted to place all the data as tag attributes, so I don't have to
     * implement this method. The empty implementation is required because the
     * XMLReaderBase defines this method as pure virtual.
     */
    virtual void CharacterDataHandler(const char *inData, int inLength) {}

    /**
     * Method for checking if settings have been read. When the xml file is empty or
     * doesn't contain the relevant tags the read operation is successful, but we don't
     * have the information we need.
     */
    bool ReadSettings() {return this->readSettings;}

  protected:
          //This is the constructor that is called by the factory to create a new object.
    XMLDataAcquisitionSettingsFileReader() : readingAcquisitionSettings(false),
                                             readingTrackerSettings(false),
                                             readSettings(false),
                                             settings(NULL) {}                                                              
  private:
    bool readingTrackerSettings;    //set to true after the tracking_system start tag is encountered
    bool readingAcquisitionSettings; //set to true after the acquisition_configuration start tag is encountered
    bool readSettings; //set to true after the first acquisition_configuration section has been read
    DataAcquisitionSettings *settings; //pointer to the DataAcquisitionSettings data structure that should be filled with the infromation from the xml file
                                        //accumulate the settings for all tools and only then update the "settings" object
    std::vector<NDIToolSettings *> tmpToolSettings;

    void ProcessDataAcquisitionSettings(const char **atts) throw (DataAcquisitionSettingsFileReaderException);
    void ProcessTrackingSystemSettings(const char **atts) throw (DataAcquisitionSettingsFileReaderException);
    void ProcessCommunicationSettings(const char **atts) throw (DataAcquisitionSettingsFileReaderException);
    void ProcessPolarisToolSettings(const char **atts) throw (DataAcquisitionSettingsFileReaderException);
    void ProcessAuroraToolSettings(const char **atts) throw (DataAcquisitionSettingsFileReaderException);

    XMLDataAcquisitionSettingsFileReader(const XMLDataAcquisitionSettingsFileReader & other);
    const XMLDataAcquisitionSettingsFileReader & operator=( const XMLDataAcquisitionSettingsFileReader & );
  };

        //set to true after the settings file has been read succesfully
  bool readSettings; 
        //set to true after the settings file name has been set  
  bool fileNameSet; 
        //name of xml file containing the settings  
  std::string settingsFileName;
        //the settings object that is filled with the data from the xml file
  DataAcquisitionSettings settings;
};


#endif //_DATA_ACQUISITION_SETTINGS_FILE_READER_H_
